"""Module for Profile and resource management ."""

# ruff: noqa: F401
from .const import (
    ResourceDevicePropertyId,
    ResourceEModeChannelPropertyId,
    ResourceEModeDevicePropertyId,
    ResourceGenericPropertyId,
    ResourceGroupObjectTablePropertyId,
    ResourceKNXNETIPPropertyId,
    ResourceLTERoutingTablePropertyId,
    ResourceObjectType,
    ResourcePollingMasterInterfacePropertyId,
    ResourceProgramPropertyId,
    ResourcePropertyId,
    ResourceRFMediumPropertyId,
    ResourceRouterPropertyId,
    ResourceSecureInterfacePropertyId,
    ResourceTextCataloguePropertyId,
)
